!***************************************
! coarsepm.f90 - coarse mesh subroutines 
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


  subroutine coarsepm
    implicit none

#ifdef DIAG
    call density_check
#endif

    call coarse_force 
       
    call force_buf

    call max_coarse_dt

    call coarse_velocity

  end subroutine coarsepm

!*******

#ifdef DIAG
  subroutine density_check
   
    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ierr
    real(kind=8) :: partsum,partot

    !$omp parallel do default(shared) private(i,j,k) reduction(+:partsum)
    do k=1,LC/NN
      do j=1,LC
        do i=1,LC
          partsum=partsum+real(rhoc(i,j,k),kind=8)
        enddo
      enddo
    enddo
    !$omp end parallel do
    
    partsum=partsum/real(LPS,kind=8)**3
    write(*,*) 'mr:',mr,'coarse density=',partsum
    call mpi_reduce(partsum,partot,1,MPI_DOUBLE_PRECISION,MPI_SUM,0, &
                    MPI_COMM_WORLD,ierr)
    if (master) write(*,*) 'average coarse grid density=',partot

  end subroutine density_check 
#endif

!*******

  subroutine coarse_force

    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ii,ir

    call filefftw(LC,rhoc,1)

    !$omp parallel do default(shared) private(i,j,k,ii,ir)
    do k=1,LC/NN
      do j=1,LC
        do i=1,LC/2+1
          ii=2*i
          ir=ii-1
          force_c(:,ir,j,k)=-rhoc(ii,j,k)*cgk(:,i,j,k)
          force_c(:,ii,j,k)=rhoc(ir,j,k)*cgk(:,i,j,k)
        enddo
      enddo
    enddo
    !$omp end parallel do

    temp_force(:,:,:)=force_c(1,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(1,:,:,1:LPSSC)=temp_force(:,:,:)

    temp_force(:,:,:)=force_c(2,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(2,:,:,1:LPSSC)=temp_force(:,:,:)

    temp_force(:,:,:)=force_c(3,:,:,1:LPSSC)
    call filefftw(LC,temp_force,-1)
    force_c(3,:,:,1:LPSSC)=temp_force(:,:,:)    

  end subroutine coarse_force

!*******

  subroutine force_buf

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'
   
    integer(kind=4) :: j
    integer(kind=4) :: status(MPI_STATUS_SIZE)
    integer(kind=4) :: tag,ierr,rank_to,rank_from,buf_size

    buf_size=3*LC*LC

    tag=42

    !$omp parallel do default(shared) private(j)
    do j=1,LC
        force_buffer(:,:,j)=force_c(:,1:LC,j,1)
    enddo
    !$omp end parallel do
    
    rank_to=mod(mr+NN-1,NN)
    rank_from=mod(mr+1,NN)

    call mpi_sendrecv_replace(force_buffer,buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr)

    force_c(:,1:LC,:,LPSSC+1)=force_buffer(:,:,:)

    tag=54
 
    !$omp parallel do default(shared) private(j) 
    do j=1,LC
        force_buffer(:,:,j)=force_c(:,1:LC,j,LPSSC)
    enddo
    !$omp end parallel do

    rank_to=mod(mr+1,NN)
    rank_from=mod(mr+NN-1,NN) 
 
    call mpi_sendrecv_replace(force_buffer,buf_size,MPI_REAL, &
        rank_to,tag,rank_from,tag,MPI_COMM_WORLD,status,ierr) 

    force_c(:,1:LC,:,0)=force_buffer(:,:,:)

  end subroutine force_buf

!*******

  subroutine max_coarse_dt

    implicit none
    include 'mpif.h'
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ierr
    real(kind=4) :: force,max_force

    max_force=0.0

    !$omp parallel do default(shared) &
    !$omp private(i,j,k,force) &
    !$omp reduction(MAX:max_force)
    do k=1,LPSSC
      do j=1,LC
        do i=1,LC
          force=force_c(1,i,j,k)*force_c(1,i,j,k)+ &
                force_c(2,i,j,k)*force_c(2,i,j,k)+ &
                force_c(3,i,j,k)*force_c(3,i,j,k)
          force=sqrt(force)
          if (force.gt.max_force) max_force=force
        enddo
      enddo
    enddo
    !$omp end parallel do

    max_force=sqrt(4.0/(max_force*aGdt/dt))
    call mpi_reduce(max_force,dt_acc_c,1,mpi_real,mpi_max,0, &
                    mpi_comm_world,ierr)
    call mpi_bcast(dt_acc_c,1,mpi_real,0,mpi_comm_world,ierr)

    if (master) write(*,*) 'maximum dt from coarse grid=',dt_acc_c

  end subroutine max_coarse_dt

!*******

  subroutine coarse_velocity

    implicit none
    include 'globalpm.fh'

    integer(kind=4) :: i,j,k,ip,im,jp,jm,kp,km,i1,i2,j2,j1,k1,k2
    integer(kind=4) :: il,jl
    real(kind=4) :: x,y,z,dx1,dx2,dy1,dy2,dz1,dz2,dV
    real(kind=4), parameter :: RLC=LC

    !$omp parallel do default(shared) &
    !$omp private(ip,i,j,k,x,y,z,i1,i2,dx1,dx2,j1,j2,dy1,dy2, &
    !$omp&              k1,k2,dz1,dz2,dV)
    do k=1,LPSSC
      do j=1,LC
        do i=1,LC
          ip=hoc(i,j,k,1)
          do
            ! use goto *gasp* instead of exit because of ifort v8
            !if (ip.eq.0) exit
            if (ip.eq.0) goto 123
            x=mod(0.25*xvp(1,ip)-0.5+RLC,RLC)
            y=mod(0.25*xvp(2,ip)-0.5+RLC,RLC)
            z=(0.25*xvp(3,ip))-0.5

            i1=floor(x)+1
            i2=mod(i1,LC)+1
            dx1=i1-x
            dx2=1-dx1
            j1=floor(y)+1
            j2=mod(j1,LC)+1
            dy1=j1-y
            dy2=1-dy1
            k1=floor(z)+1
            k2=k1+1
            dz1=k1-z
            dz2=1-dz1

            dV=aGdtC*dx1*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j1,k1)*dV
            dV=aGdtC*dx2*dy1*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j1,k1)*dV
            dV=aGdtC*dx1*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j2,k1)*dV
            dV=aGdtC*dx2*dy2*dz1
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j2,k1)*dV
            dV=aGdtC*dx1*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j1,k2)*dV
            dV=aGdtC*dx2*dy1*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j1,k2)*dV
            dV=aGdtC*dx1*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i1,j2,k2)*dV
            dV=aGdtC*dx2*dy2*dz2
            xvp(4:6,ip)=xvp(4:6,ip)+force_c(:,i2,j2,k2)*dV

            ip=ll(ip)
          enddo
 123      continue
        enddo
      enddo
    enddo
    !$omp end parallel do

  end subroutine coarse_velocity
