!! This is the fortran wrapper file for ipp fft routines
!! Limitations:    
!! 1) This only works with cubic arrays (nx=ny=nz)
!! 2) This cannot be threaded internally, however multiple threads
!!    can call these routines simultaneously.
!! 3) The size of the arrays which are to be transformed cannot
!!    change during the duration of the program

!------------------------------------------------------------- 
! IPP 1d fft routine 
! parameters: 
!            a - the input real array 
!  n - size of the array 
!            isign - the direction of the transform 
!  1 - forward, -1 - backwards 

subroutine fft1dipp(a,n,isign)
   implicit none 
   integer n,isign,i 
   real a(n+2)

   call fft1dippsrcinit(n)
   call fft1dippsrc(a,a,isign)
end subroutine fft1dipp

!------------------------------------------------------------
! IPP 2d fft routine
! parameters:
!             a - the input(or output, for isign=-1) real array
!                          dim: (nx+2,ny)
!             c - the output(or input, for isign=-1) complex array
!                          dim: (ny,nx/2+1)
!             n - the size
!             isign - the direction (1 - forward, -1 - backwards)

subroutine fft2(a,nx,ny,isign)
complex, dimension(ny,nx/2+1) :: a
complex, dimension(nx/2+1,ny) :: c

if (isign .eq. -1) c=transpose(a)
call fft2dipp(c,a,c,c,nx,ny,isign)
if (isign .eq. 1) a=transpose(c)

end

recursive subroutine fft2dipp(c,a,b1,d1,nx,ny,isign)
   implicit none
   integer nx,ny,isign,i,j,jo,jmax
   integer, parameter:: nblock=16,nblock2=nblock*2
   complex, dimension(ny,nx/2+1), target :: a,b1
   complex, dimension(nx/2+1,ny), target :: c,d1
   complex, dimension(:,:), pointer :: b,d
   logical inplace
   
   inplace=%loc(c).eq.%loc(a)
   if (inplace) then
     b=>b1
     d=>d1
   else
     b=>a
     d=>c
   endif
   if (isign.eq.1) then   
     do jo=1,ny,nblock2
        jmax=min(ny,jo+nblock2-1)
        do j=jo,jmax
           call fft1dippsrc(c(1,j),c(1,j),isign)
        enddo
        call transpose2(b,c,nx/2+1,ny,jo,jmax)
     enddo
     do j=1,nx/2+1
        call fft1dippscc(b(1,j),a(1,j),isign)
     enddo
   else
     do jo=1,nx/2+1,nblock
        jmax=min(nx/2+1,jo+nblock-1)
        do j=jo,jmax
           call fft1dippscc(a(1,j),a(1,j),isign)
        enddo
        call transpose2(d,a,ny,nx/2+1,jo,jmax)
     enddo
     do j=1,ny
        call fft1dippsrc(d(1,j),c(1,j),isign)
     enddo
   endif

end subroutine fft2dipp

!-------------------------------------------------------------
! IPP 3d fft routine
! parameters:
!             a - the input(or output, depending on direction) real array
!                           dim: (n+2,n,n)
!             c - the output(or input, depending on direction) complex array
!                           dim: (n,n/2+1,n)
!             n - the size
!             isign - the direction (1 - forward, -1 - backwards)

recursive subroutine fft3dipp(a,c,nx,ny,nz,isign)
   implicit none
   integer nx,ny,nz, isign, i, j, k
   real a(nx+2,ny,nz)
   complex  c(nz,nx/2+1,ny)
   real tmp((nx+2)*ny)
   character*40 dt
   logical firsttime
   integer nlast
   common /ipp3d/ nlast,firsttime
   data firsttime /.true./

   if (firsttime) then
      firsttime = .false.
      nlast=nx
      call fft1dippsrcinit(nx)
      call fft1dippsccinit(nz)
   endif

   if (nx .ne. nlast .or. ny .ne. nlast .or. nz .ne. nlast) then
      write(*,*) 'ippfft3d: parameters changes',nx,ny,nz,nlast
      stop
   endif

   if (isign .eq. 1) then

         do i=1,nz
            call fft2dipp(a(1,1,i),c(1,1,i),tmp,tmp,nx,ny,isign)
         enddo
         call date_and_time(time=dt)
         call transpose3f(c,nz,nx/2+1,ny)
         call date_and_time(time=dt)
    else

       call transpose3(c,nz,nx/2+1,ny)

         call date_and_time(time=dt)
         do i=1,nz
            call fft2dipp(a(1,1,i),c(1,1,i),tmp,tmp,nx,ny,isign)
         enddo
    endif

end subroutine fft3dipp

recursive subroutine transposecc0(dest,src,nx,ny)
implicit none
integer nx,ny
complex src(nx,ny),dest(ny,nx)

call ipptransposes(dest,src,nx,ny)
return
end 

recursive subroutine transpose2(dest,src,nx,ny,jmin,jmax)
implicit none
integer nx,ny,i,j,io,jo,jmin,jmax
integer, parameter :: nblock=8
integer(kind=8) :: src(nx,ny),dest(ny,nx)

!!$omp parallel do default(shared) private(io,j,i)
do jo=jmin,jmax,nblock
   do io=1,nx,nblock
      do j=jo,min(ny,jo+nblock-1)
         do i=io,min(nx,io+nblock-1)
            dest(j,i)=src(i,j)
         end do
      end do
   end do
end do
end


recursive subroutine transposecc1(dest,src,nx,ny)
implicit none
integer nx,ny,i,j,io,jo
integer, parameter :: nblock=16
complex src(nx,ny),dest(ny,nx)

!dest=transpose(src)
!return
!$omp parallel do default(shared) private(io,j,i)
do jo=1,ny,nblock
   do io=1,nx,nblock
      do j=jo,min(ny,jo+nblock-1)
         do i=io,min(nx,io+nblock-1)
            dest(j,i)=src(i,j)
         end do
      end do
   end do
end do
end


recursive subroutine transposecc(dest,src,nx,ny)
implicit none
integer nx,ny,i,j,io,jo
integer, parameter :: nblock=16
integer(kind=8) :: src(nx,ny),dest(ny,nx),tmp(nx,ny)
integer(kind=8) :: tt

if (%loc(dest) .ne. %loc(src) ) then
write(*,*) 'dest != src'
call transposecc1(dest,src,nx,ny)
return
endif
call ippcopy8(tmp,src,nx*ny)
!$omp parallel do default(shared) private(io,j,i)
do jo=1,ny,nblock
   do io=1,nx,nblock
      do j=jo,min(ny,jo+nblock-1)
         do i=io,min(nx,io+nblock-1)
            dest(j,i)=tmp(i,j)
         end do
      end do
   end do
end do
end

recursive subroutine transpose3(dest,nx,ny,nz)
implicit none
integer nx,ny,nz,i,j,k,io,ko
integer, parameter :: nblock=16
integer(kind=8) :: dest(nx,ny,nz)
integer(kind=8) :: tt

if (nx .ne. nz) write(*,*) 'transpose3:nx,ny=',nx,ny
!$omp parallel do default(shared) private(ko,k,i,tt,io)
do j=1,ny
  do ko=1,nz,nblock
   do k=ko,min(nz,ko+nblock-1)
     call fft1dippscc(dest(1,j,k),dest(1,j,k),-1)
   enddo
   do k=ko,min(nz,ko+nblock-1)
      do i=min(nz,ko+nblock-1),k+1,-1
            tt=dest(i,j,k)
            dest(i,j,k)=dest(k,j,i)
            dest(k,j,i)=tt
      end do
   end do
   do io=ko-1,1,-nblock
      do k=ko,min(nz,ko+nblock-1)
         do i=io,max(1,io-nblock+1),-1
            tt=dest(i,j,k)
            dest(i,j,k)=dest(k,j,i)
            dest(k,j,i)=tt
         end do
      end do
   end do
  end do
end do
end

recursive subroutine transpose3f(dest,nx,ny,nz)
implicit none
integer nx,ny,nz,i,j,k,io,ko
integer, parameter :: nblock=8
integer(kind=8) :: dest(nx,ny,nz)
integer(kind=8) :: tt

if (nx .ne. nz) write(*,*) 'transpose3:nx,ny=',nx,ny
!$omp parallel do default(shared) private(ko,k,i,tt,io)
do j=1,ny
  do ko=1,nz,nblock
   do k=ko,min(nz,ko+nblock-1)
      do i=k+1,min(nx,ko+nblock-1)
            tt=dest(i,j,k)
            dest(i,j,k)=dest(k,j,i)
            dest(k,j,i)=tt
      end do
   end do
   do io=ko+nblock,nx,nblock
      do k=ko,min(nz,ko+nblock-1)
         do i=io,min(nx,io+nblock-1)
            tt=dest(i,j,k)
            dest(i,j,k)=dest(k,j,i)
            dest(k,j,i)=tt
         end do
      end do
   end do
   do k=ko,min(nz,ko+nblock-1)
      call fft1dippscc(dest(1,j,k),dest(1,j,k),1)
   end do
  end do
end do
end
