!***********************************************************
! pmfast.f90 - 2 level MPI/OpenMP Cosmological N-body Solver 
! Copyright (C) 2004 Hugh Merz - merz@cita.utoronto.ca
! latest revision Feb 11 2004
!***********************************************************
!This program is free software; you can redistribute it and/or
!modify it under the terms of the GNU General Public License
!as published by the Free Software Foundation; either version 2
!of the License, or (at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

program pmfast

  implicit none
  include 'globalpm.fh' 
  include 'mpif.h'
 
  integer(kind=4) :: ierr

if (pairtest) then
  call initialize_pairtest
  do
    call init_pair_vars
    call set_pair
    call pairtest_mesh
    call measure_pair
    if (nts.ge.MAX_NTS) exit
  enddo
else
  call initialize_pm    
  call particle_count

  do
    call multistep
    call particle_mesh
    call particle_count
    call mpi_barrier(mpi_comm_world,ierr)
    if (nts.ge.MAX_NTS) then
      if (master) write(*,*) 'exceeded timestep limit'
      exit
    endif
    if (check_step) then
      dt_old=0.0
      call update_position
      call checkpoint
      if (stopsim) exit
    endif
    if (proj_step) then
      dt_old=0.0
      call update_position
      call density_proj
    endif
  enddo
endif

  call filefftw_finalize 

  if (master) write(*,*) 'finished main loop of program:', t_elapsed(wcts)

  call mpi_finalize(ierr)

contains

!*******

  subroutine particle_mesh
    implicit none

    integer(kind=4) :: i,j,k,ip,in,ipp
    integer(kind=4) :: isec,jsec
    integer(kind=4) :: tloc
    real(kind=8) :: fdentemp,fdentot
    real(kind=4) :: maxf

    integer(kind=4) :: thrdnum
    integer(kind=4) :: omp_get_thread_num 
    external omp_get_thread_num

    call update_position
 
    if (master) write(*,*) 'finished position update:', t_elapsed(wcts)

    if (nts.eq.1.or.restart_step) then
      call link_list
      restart_step=.false.
      if (master) write(*,*) 'finished initial linked list:', t_elapsed(wcts)
    endif

    call particle_pass

    if (master) write(*,*) 'finished particle pass:', t_elapsed(wcts)

    call link_list

    if (master) write(*,*) 'finished constructing linked list:', t_elapsed(wcts)

    fdentemp=0.0
    fdentot=0.0
    if (coarse_step) rhoc(:,:,:)=0.0

    !$omp parallel default(shared) &
    !$omp private(jsec,isec,thrdnum) &
    !$omp reduction(+:fdentemp)
    thrdnum=1
    thrdnum=omp_get_thread_num()+1
    fmax(thrdnum)=0.0
    fmaxloc(:,thrdnum)=0
    !$omp do schedule (static,1)
    do jsec=1,NN
       do isec=1,NN
         if (fine) call finepm(isec,jsec,thrdnum,fdentemp)
       enddo
    enddo
    !$omp end do
    !$omp end parallel

    maxf=maxval(fmax)
    tloc=maxloc(fmax,1)

#ifdef DIAG
    do i=0,NN-1
      if (mr.eq.i) then
        write(*,*) '!!! maxf =',mr,fmaxloc(1:3,tloc),maxf
      endif
      call mpi_barrier(mpi_comm_world,ierr)
    enddo
#endif

    call mpi_reduce(maxf,dt_acc,1,mpi_real,mpi_max,0, &
                    mpi_comm_world,ierr)

    if (master) then
      dt_acc=1.0/sqrt(dt_acc*aGdt/dt)
      write(*,*) 'maximum timestep from fine velocity=',dt_acc
    endif

    call mpi_bcast(dt_acc,1,mpi_real,0,mpi_comm_world,ierr)
 
   if (master) write(*,*) 'finished fine mesh:', t_elapsed(wcts)

#ifdef DIAG
    call mpi_reduce(fdentemp,fdentot,1,mpi_double_precision,mpi_sum,0, &
         mpi_comm_world,ierr)

    write(*,*)'fine grid density mr',mr,fdentemp
    fdentot=fdentot/real(RLPS,kind=8)**3
    if (master) write(*,*) 'average fine grid density =',fdentot
#endif

    if (COARSE.eq.0) then
      if (master) write(*,*) 'skipped coarse grid'
    else 
      if (coarse_step) call coarsepm 
    endif

    if (master) write(*,*) 'finished coarse mesh:', t_elapsed(wcts)

    nploc=nploc-ngloc
 
  end subroutine particle_mesh

!*******

  subroutine particle_count
    implicit none

    integer(kind=4) :: pc_status(MPI_STATUS_SIZE)
    integer(kind=4) :: pc_tag
    integer(kind=4) :: i
    integer(kind=8) :: nptemp,nploc8,pc_nptot

    nploc8=int(nploc,kind=8)

    do i=0,NN-1
      if (mr.eq.i) write(*,*) 'mr:',mr,'number of particles =',nploc8
      call mpi_barrier(MPI_COMM_WORLD,ierr)
    enddo

    do i=0,NN-1
      pc_tag=i
      if (i.eq.0) then
        pc_nptot=nploc8
      else
        if (mr.eq.0) then
          nptemp=0
          call mpi_recv(nptemp,1,MPI_2INTEGER,i,pc_tag, &
                        MPI_COMM_WORLD,pc_status,ierr)
          pc_nptot=pc_nptot+nptemp
        elseif (mr.eq.i) then
          call mpi_send(nploc8,1,MPI_2INTEGER,0,pc_tag, &
                        MPI_COMM_WORLD,ierr)
        endif
        call mpi_barrier(MPI_COMM_WORLD,ierr)
      endif
    enddo  

    if (master) write(*,*) 'total particles =',pc_nptot

  end subroutine particle_count

!*******

  subroutine checkpoint
    implicit none

    character (len=MSL) :: ofile
    character (len=MSL) :: ifile
    integer(kind=4) :: i
    integer(kind=4) :: blocksize
    integer(kind=4) :: num_writes,nplow,nphigh
    real(kind=4) :: z_write

    if (master) z_write=z_cp(cur_cp)
    call mpi_bcast(z_write,1,mpi_real,0,mpi_comm_world,ierr)

    blocksize=(2047*1024*1024)/24
    num_writes=nploc/blocksize+1
 
    write(ifile,'("xvp",i1,".dat")') mr
    if (mod(cur_cp,2).eq.1) then
      ofile=OUT1//ifile
    else
      ofile=OUT2//ifile
    endif
    open (unit=12,file=ofile,status='replace',form='binary')
    write(12) nploc
    do i=1,num_writes
      nplow=(i-1)*blocksize+1
      nphigh=min(i*blocksize,nploc)
      write(12) xvp(1:6,nplow:nphigh) 
    enddo

    close(12)

    cur_cp=cur_cp+1

    if (master) then
      write(ifile,'(f5.1,"params.dat")') z_write
      ifile=adjustl(ifile)
      ofile=OUT1//ifile
      open (unit=14,file=ofile,form='binary')
      write(14) a,t,nts,dt_acc,dt_acc_c,cur_cp,cur_rp
      close(14)
    endif

    write(*,*) 'Called checkpoint - mr:',mr

  end subroutine checkpoint

!*******

  subroutine update_position
    implicit none
   
    integer(kind=4) :: ip
    real(kind=4) :: dttot

      dttot=0.5*(dt+dt_old)

      !$omp parallel do default(shared) private(ip)
      do ip=1,nploc
        xvp(1:2,ip)=modulo(xvp(1:2,ip)+xvp(4:5,ip)*dttot,RLPS)
        xvp(3,ip)=xvp(3,ip)+xvp(6,ip)*dttot
      enddo
      !$omp end parallel do
  
  end subroutine update_position

!********

end program pmfast
